from .exception import WrongOrderError
from PIL import ImageTk
from queue import Empty

class Video():
    def __init__(self, stream, mode="w"):
        self.ffmpeg = stream
        self.mode=mode
        self.played=False
        self.stopwatch=self.ffmpeg.stopwatch
        self.state="stop"
        self.old_frame=None
        self.old_frame2=None
        self.tkimage=None
    def play(self, frame, frame_type):
        if not self.ffmpeg.loader["loaded"]: raise WrongOrderError("Stream is not loaded.")
        info=self.ffmpeg.info["streams"]["video"][self.ffmpeg.loader["video"]]
        self.frame=frame
        self.frame_type=frame_type
        self.state="play"
        self.frame.after(info["start_time"], self._play)
        self.played=True
    def _play(self):
        if self.state=="play":
            try:
                frame=self.ffmpeg._videoQ.get_nowait()
            except Empty:
                self.frame.after(10, self._play)
            else:
                info=self.ffmpeg.info["streams"]["video"][self.ffmpeg.loader["video"]]
                sleep_time=1/info["fps"]
                gap=self.stopwatch.getTime()-frame[0]
                border=0.5
                if gap<float(0):
                    self.frame.after(0, self._play)
                elif gap>float(border):
                    self.frame.after(int(gap), self._play)
                elif gap<=float(border):
                    if not self.old_frame is None:
                        self.old_frame.close()
                        self.old_frame2=self.tkimage
                    self.tkimage=ImageTk.PhotoImage(frame[1], master=self.frame)
                    if self.frame_type=="label":
                        self.frame.configure(image=self.tkimage)
                    self.old_frame=frame[1]
                    self.frame.after(int(sleep_time*1000), self._play)
        elif self.state == "pause":
            pass
        elif self.state == "stop":
            pass
    def pause(self):
        self.state = "pause"
    def resume(self):
        self.state = "play"
        self.frame.after(0, self._play)
    def stop(self):
        self.state="stop"
    def close(self):
        self.state="stop"


import av, threading, time, numpy, warnings
from av.filter import Graph
from  av.audio.fifo import AudioFifo 
from queue import Full, Queue, Empty
import av.datasets
from .exception import MediaFileError
from .lib import StopWatch

def toImage(frame):
    try:
        return (frame.time, frame.to_image())
    except:
        return (frame.time, frame.to_rgb().to_image())
def toSdArray(frame):
    return (frame.time, numpy.transpose(frame.to_ndarray()).copy(order='C'))

class Filter():
    def __init__(self, stream=None, width=None, height=None, format=None, name=None, audio=False):
        self.filter=Graph()
        self.src=self.filter.add_buffer(template=stream, width=width, height=height, format=format, name=name)
        self.audio=audio
    def addFilter(self, filter, arg):
        f = self.filter.add(filter=filter, args=arg)
        self.src.link_to(f)
        f.link_to(self.filter.add("buffersink"))#thinking...
        self.filter.configure()
    def Process(self, frame):
        self.filter.push(frame=frame)
        return self.filter.pull()
    def getFilter(self):
        pass
    def availableFilters(self):
        return av.filter.filters_available

class Stream():
    def __init__(self, path, mode="r", **options):#TODO: write support
        self.stopwatch=StopWatch(error=False)
        self.ffmpeg = av.open(av.datasets.curated(path), mode=mode, **options)
        self.info = {
            "codec_name" : self.ffmpeg.format.name.split(","),
            "codec_long" : self.ffmpeg.format.long_name,
            "mode" : mode
        }
        self.streams = self.ffmpeg.streams
        if mode == "r":
            self.info.update(bit_rate=self.ffmpeg.bit_rate)
            self.info.update(duration=self.ffmpeg.duration)
            self.info.update(size=self.ffmpeg.size)
            self.info.update(start_time=self.ffmpeg.start_time)
            def _stream_info(streams):
                def toDict(st):
                    if st.type == "video" or st.type == "audio":
                        codec=st.codec_context
                        o = {"duration":st.duration, "id":st.id, "index":st.index, "lang":st.language, "meta":st.metadata,
                        "frame_count":st.frames, "bit_rate":codec.bit_rate, "codec_name":codec.name, "codec_long":codec.codec.long_name,
                        "delay":codec.codec.delay, "time_base":st.time_base, "start_time":st.start_time}
                        if st.type == "audio":
                            o.update(channel=codec.channels, channel_name=codec.layout.name, sample_rate=st.sample_rate, frame_size=codec.frame_size)
                        elif st.type == "video":
                            o.update(height=codec.height,width=codec.width, fps=st.base_rate, aspect_ratio=codec.sample_aspect_ratio)
                        return o
                return [toDict(st) for st in streams]
            self.info.update(streams={
                "video" : _stream_info(self.streams.video),
                "audio" : _stream_info(self.streams.audio),
                "data" : _stream_info(self.streams.data),
                "subtitles" : _stream_info(self.streams.subtitles),
                "other" : _stream_info(self.streams.other)
            })
            self.loader={"state":"stop", "thread":None, "queue_min":100, "queue_max":300, "audio":None, "video":None, "loaded":False, "audio_processor":[toSdArray], "video_processor":[toImage], "frame_size": 0}
        elif mode == "w":
            pass
    def load(self, audio=None, video=None, queue_min=None, queue_max=None, wait=True):
        if not queue_min is None:
            self.loader["queue_min"]=queue_min
        if not queue_max is None:
            self.loader["queue_max"]=queue_max
        mux_source=[]
        self._oldFrames=[]
        if not audio is None:
            if len(self.info["streams"]["audio"]) == 0:
                raise MediaFileError("This File doesn't contain audio.")
            if len(self.info["streams"]["audio"]) < audio:
                raise MediaFileError(f"Audio stream '{audio}' is not exists.")
            self.loader["audio"]=audio
            self._audioPreQ=AudioFifo()
            self._audioQ=Queue(maxsize=self.loader["queue_max"])
            self.loader["frame_size"]=self.info["streams"]["audio"][audio]["frame_size"]
            mux_source.append(self.streams.get(audio=audio))
        if not video is None:
            if len(self.info["streams"]["video"]) == 0:
                raise MediaFileError("This File doesn't contain video.")
            if len(self.info["streams"]["video"]) < video:
                raise MediaFileError(f"Video stream '{video}' is not exists.")
            self.loader["video"]=video
            self._videoQ=Queue(maxsize=self.loader["queue_max"])
            mux_source.append(self.streams.get(video=video))
        if audio is None and video is None:
            return
        if not audio is None and not video is None:
            self.loader["generator"]=[self.ffmpeg.demux(audio=audio, video=video)]
        elif not audio is None:
            self.loader["generator"]=[self.ffmpeg.decode(mux_source[0])]
        elif not video is None:
            self.loader["generator"]=[self.ffmpeg.decode(mux_source[0])]
        self.loader["state"]="load"
        self.loader["thread"]=threading.Thread(target=self._loader)
        self.loader["thread"].start()
        if wait:
            while True:
                if self.loader["loaded"]:
                    break
                time.sleep(0.1)
        self.stopwatch.start()
    def _loader(self):
        for generator in self.loader["generator"]:
            for frame in generator:
                if self.loader["state"] == "stop":
                    break
                audio=self._audioQ.qsize()>= self.loader["queue_min"] if hasattr(self, "_audioQ")  else True
                video=self._videoQ.qsize()>= self.loader["queue_min"] if hasattr(self, "_videoQ")  else True
                if audio and video:
                    self.loader["loaded"]=True
                else:
                    self.loader["loaded"]=False
                req=[]
                while len(req)==0:
                    if self.loader["state"] == "pause":
                        time.sleep(0.001)
                        continue
                    elif self.loader["state"] == "stop":
                        break
                    if not self.loader["audio"] is None and self._audioQ.qsize() < self.loader["queue_min"]:
                        req.append("a")
                    if not self.loader["video"] is None and self._videoQ.qsize() < self.loader["queue_min"]:
                        req.append("v")
                    time.sleep(0.001)
                if isinstance(frame, av.packet.Packet):
                    frames=frame.decode()
                else:
                    frames=[frame]
                for frame in frames:
                    if isinstance(frame, av.audio.AudioFrame):
                        try:
                            self._audioPreQ.write(frame)
                        except:
                            pass
                        else:
                            frame=self._audioPreQ.read(self.loader["frame_size"])
                        if not frame is None:
                            if not frame in self._oldFrames:
                                self._oldFrames.append(frame)
                            for processor in self.loader["audio_processor"]:
                                frame=processor(frame)
                            try:
                                self._audioQ.put_nowait(frame)
                            except Full:
                                warnings.warn("Can't put frame to audio queue.(Queue is full)", Warning)
                    elif isinstance(frame, av.video.VideoFrame):
                        if not frame in self._oldFrames:
                            self._oldFrames.append(frame)
                        for processor in self.loader["video_processor"]:
                            frame=processor(frame)
                        try:
                            self._videoQ.put_nowait(frame)
                        except Full:
                            warnings.warn("Can't put frame to video queue.(Queue is full)", Warning)
                    else:
                        warnings.warn("Unknown frame type.", Warning)
            else:
                self.loader["state"]="stop"
                return
    def seek(self, point):
        self.stop(block=True)
        self.clear()
        for i, f in enumerate(self._oldFrames):
            if f.time > point:
                break
        self.loader["generator"]=[self._oldFrames[i:], self.loader["generator"][-1]]
        self.stopwatch.setTime(point)
        self.loader["state"]="load"
        self.loader["loaded"]=False
        self.loader["thread"]=threading.Thread(target=self._loader)
        self.loader["thread"].start()
        while True:
            if self.loader["loaded"]:
                break
            time.sleep(0.1)
        self.stopwatch.start()
    def clear(self):
        if hasattr(self, "_audioPreQ"):
            self._audioPreQ.read(0)
        if hasattr(self, "_audioQ"):
            while 1:
                try:
                    self._audioQ.get_nowait()
                except Empty:
                    break
        if hasattr(self, "_videoQ"):
            while 1:
                try:
                    self._videoQ.get_nowait()
                except Empty:
                    break
    def pause(self):
        self.loader["state"]="pause"
        self.stopwatch.stop()
    def resume(self):
        self.loader["state"]="load"
        self.stopwatch.start()
    def stop(self, block=True):
        self.loader["state"]="stop"
        if block: self.loader["thread"].join()
        self.stopwatch.stop()
    def close(self):
        self.stopwatch.stop()
        self.ffmpeg.close()


import time
try:
    from .exception import WrongOrderError
except:
    from exception import WrongOrderError

class StopWatch():
    def __init__(self, error=True):
        self.start_time = 0
        self.stop_time = 0
        self.time = 0
        self.isStop = True
        self.error = error
    def start(self):
        if self.isStop:
            self.start_time = time.time() 
            self.isStop=False
        else:
            if self.error: raise WrongOrderError("Already Started.")
    def stop(self):
        if self.isStop:
            if self.error: raise WrongOrderError("Already Stoped.")
        else:
            self.time+=time.time()-self.start_time
            self.isStop=True
    def clear(self):
        self.time = 0
        self.stop_time=0
        self.isStop=True
    def getTime(self):
        if self.isStop:
            return self.time
        else:
            return self.time+(time.time()-self.start_time)
    def setTime(self, t):
        stoped=self.isStop
        try:
            self.stop()
        except:
            pass
        self.time=t
        if not stoped:
            self.start()


# -*- coding: utf-8 -*-
from os import path
import setuptools

root_dir = path.abspath(path.dirname(__file__))

def _requirements():
    return [name.rstrip() for name in open(path.join(root_dir, 'requirements.txt')).readlines()]

with open(path.join(root_dir, 'requirements.txt'), "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="tkmedia",
    version="0.1.1",
    author="Marusoftware",
    author_email="marusoftware@outlook.jp",
    description="Tkinter Media Support.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Marusoftware/tkmedia3",
    project_urls={
        "Bug Tracker": "https://github.com/Marusoftware/tkmedia3/issues",
        "Documents": "https://marusoftware.github.io/tkmedia3/"
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
    packages=["media"],
    license="MIT",
    python_requires=">=3.8",
    install_requires=_requirements(),
    keywords='tkmedia',
)


import sounddevice, queue
from .exception import ModeError, WrongOrderError
from .lib import StopWatch

def getDevices(device=None, kind=None):
    return list(sounddevice.query_devices(device=device, kind=kind))
def getVersion():
    return sounddevice.get_portaudio_version()
def getHostApi(index=None):
    return sounddevice.query_hostapis(index=index)

class Sounddevice():
    def __init__(self, mode, streamOptions={}, dataQueue=None, rw_autoStop=True, stopwatch=StopWatch(error=False)):
        self.mode = mode
        if self.mode=="rw":
            if not "callback" in streamOptions: streamOptions.update(callback=self._Queue2)
            self.sdStream= sounddevice.Stream(dtype="float32", **streamOptions)
            self.rw_autoStop=rw_autoStop
        elif self.mode == "r":
            if not "callback" in streamOptions: streamOptions.update(callback=self._Queue)
            self.sdStream= sounddevice.InputStream(dtype="float32", **streamOptions)
        elif self.mode == "w":
            if not "callback" in streamOptions: streamOptions.update(callback=self._Queue)
            self.sdStream= sounddevice.OutputStream(dtype="float32", **streamOptions)
        else:
            raise ModeError("Unknown mode.")
        if dataQueue is None:
            if self.mode == "rw":
                self.dataQueue = (queue.Queue(), queue.Queue())
            else:
                self.dataQueue = queue.Queue()
        else:
            self.dataQueue=dataQueue
        self.state="stop"
        self.stopwatch=stopwatch
        self.last_frametime=0
    def fixDataQueue(self, dataQueue=None):
        if dataQueue is None:
            if self.mode == "rw":
                self.dataQueue = (queue.Queue(), queue.Queue())
            else:
                self.dataQueue = queue.Queue()
        else:
            self.dataQueue=dataQueue
    def Record(self):
        if self.mode == "w": raise ModeError("Can't record in write Mode.")
        if self.mode == "rw" and self.state == "play": self.state = "both"
        else: self.state="rec"
        self.sdStream.start()
    def Play(self):
        if self.mode == "r": raise ModeError("Can't play in read Mode.")
        if self.state in ["both", "play", "rec"]:
            raise WrongOrderError("Alredy Played.")
        if self.mode == "rw" and self.state == "rec": self.state = "both"
        else: self.state="play"
        self.sdStream.start()
    def Stop(self):
        self.state="stop"
    def Pause(self):
        self.state="pause"
    def Resume(self):
        if self.state != "pause": WrongOrderError("Not Paused")
        self.state="play"
    def Close(self):
        self.sdStream.close()
    def _Queue(self, data, frames, time, status):
        if self.state == "play":
            try:
                gap=self.stopwatch.getTime()-self.last_frametime
                border=0.5
                if gap<float(0):
                    print("\r minus", self.stopwatch.getTime(), self.last_frametime, end="")
                    data.fill(0)
                    self.last_frametime=self.stopwatch.getTime()
                else:
                    frame_time, datafQ=self.dataQueue.get_nowait()
                    if gap>float(border):
                        frame_time, datafQ=self.dataQueue.get_nowait()
                        data[:] = datafQ
                    else:
                        data[:] = datafQ
                    status.output_underflow=False
                    self.last_frametime=frame_time
            except queue.Empty:
                print("\r empty", self.dataQueue, end="")
                status.output_underflow=True
                data.fill(0)
        elif self.state == "rec":
            try:
                self.dataQueue.put_nowait(data)
                status.input_overflow=False
            except queue.Full:
                status.input_overflow=True
        elif self.state == "stop":
            raise sounddevice.CallbackAbort
        elif self.state == "pause":
            data.fill(0)
    def _Queue2(self, indata, outdata, frames, time, status):
        if self.state == "both":
            try:
                self.dataQueue[0].put_nowait(indata)
                status.input_overflow=False
            except queue.Full:
                status.input_overflow=True
            try:
                outdata[:] = self.dataQueue[1].get_nowait()
                status.output_underflow=False
            except:
                status.output_underflow=True
        elif self.state == "stop":
            raise sounddevice.CallbackAbort
        elif self.state == "pause":
            outdata.fill(0)


from .audio import Audio
from .video import Video
from .picture import Picture
from .exception import MediaFileError, ModeError, WrongOrderError
from ._ffmpeg import Stream, Filter
from .lib import StopWatch

__version__="0.1.1"
__license__="MIT License"
__author__="Marusoftware"
__author_email__="marusoftware@outlook.jp"
__url__="https://marusoftware.net"
__all__=["Video", "Audio", "Picture", "Media", "Filter"]

class Media():
    def __init__(self, url, mode="r", **options):
        """
        Initialize media library.
        url(str or file-like object): Media file url. This will pass to ffmpeg.
        mode(str): Opening mode. You can set to 'r' or 'w'.
        **options: Options for PyAV.
        """
        self.ffmpeg=Stream(url, mode, **options)
        self.info=self.ffmpeg.info
        self.ffmpegs=self.ffmpeg.streams
        self.stopwatch=self.ffmpeg.stopwatch
        self.status="pause"
    def Play(self, audioDevice=None, videoFrame=None, video=None, audio=None, frame_type="label"):
        """
        Play media.
        audioDevice(int): An number of Device(You can get with media.audio.getDevices func)
        videoFrame(tkinter.Label or tkinter.Canvas): An frame to playing video
        frame_type(str): videoFrame type(Label->label, Canvas->canvas)
        video(int): video stream number
        audio(int): audio stream number
        """
        self.ffmpeg.load(audio=audio, video=video, wait=True)
        if not video is None:
            self.video=Video(self.ffmpeg, mode="w")
        if not audio is None:
            self.audio=Audio(self.ffmpeg, mode="w")
        if not audio is None:
            self.audio.play(device=audioDevice)
        if not video is None:
            self.video.play(frame=videoFrame, frame_type=frame_type)
        self.play_options={"audioDevice":audioDevice, "videoFrame":videoFrame, "video":video, "audio":audio, "frame_type":frame_type}
        self.status="play"
    def Pause(self):
        """
        Pause playing.
        """
        if self.status != "play": raise WrongOrderError("Media is not playing.")
        self.ffmpeg.pause()
        try:
            self.video.pause()
        except: pass
        try:
            self.audio.pause()
        except: pass
        self.status="pause"
    def Resume(self):
        """
        Resume playing.
        """
        if self.status != "pause": raise WrongOrderError("Media is not paused.")
        self.ffmpeg.resume()
        try:
            self.video.resume()
        except: pass
        try:
            self.audio.resume()
        except: pass
        self.status="play"
    def Seek(self, point):
        """
        Seek media.
        """
        self.Pause()
        self.ffmpeg.seek(point)
        self.Resume()
        self.status="play"
    def Stop(self):
        """
        Stop media.
        """
        self.ffmpeg.stop()
        try:
            self.video.stop()
            pass
        except: pass
        try:
            self.audio.stop()
        except: pass
        self.status="stop"
    def Close(self):
        """
        Close media.
        """
        if self.status !="stop": self.Stop()
        try:
            self.audio.close()
        except: pass
        try:
            self.video.close()
        except: pass
        self.ffmpeg.close()
        self.status="close"


from ._Sounddevice import Sounddevice, getDevices, getHostApi, getVersion
from .exception import WrongOrderError

class Audio():
    def __init__(self, stream, mode="w"):
        self.ffmpeg = stream
        self.mode=mode
        self.played=False
        self.stopwatch=self.ffmpeg.stopwatch
    def play(self, channels=2, device=None):
        if not self.ffmpeg.loader["loaded"]: raise WrongOrderError("Stream is not loaded.")
        if self.played:
            self.sdStream.Stop()
        info=self.ffmpeg.info["streams"]["audio"][self.ffmpeg.loader["audio"]]
        self.channels=channels
        self.device=device
        streamOptions={"frames_per_buffer":info["frame_size"]}
        if self.mode == "rw":
            if type(self.device) in [list, tuple]:
                streamOptions.update(input_device_index=self.device[0], output_device_index=self.device[1])
            else:
                streamOptions.update(input_device_index=self.device, output_device_index=self.device)
        elif self.mode == "r":
            streamOptions.update(input_device_index=self.device)
        else:
            streamOptions.update(output_device_index=self.device)
        self.streamOptions=streamOptions
        self.sdStream=Sounddevice(mode=self.mode, dataQueue=self.ffmpeg._audioQ, stopwatch=self.stopwatch, streamOptions={"samplerate":info["sample_rate"], "blocksize":info["frame_size"], "channels":self.channels, "device":self.device})
        self.played=True
        self.sdStream.Play()
    def pause(self):
        if not self.played: raise WrongOrderError("Not played")
        self.sdStream.Pause()
    def resume(self):
        if not self.played: raise WrongOrderError("Not played")
        self.sdStream.fixDataQueue(self.ffmpeg._audioQ)
        self.sdStream.Resume()
    def stop(self):
        if not self.played: raise WrongOrderError("Not played")
        self.sdStream.Stop()
    def close(self):
        self.sdStream.Close()